"""JSON APIs

This module contains functions for sending HTTP requests & processing JSON responses.

It can be used to integrate PPA tasks with systems & applications that have JSON APIs.
"""
import json
import requests
import logging

from typing import Dict, Optional, Union, List, Any

from typeguard import check_type

from . import exceptions
from .records import JsonResponse, BasicAuth


logger = logging.getLogger(__name__)


Kwargs = Dict[str, Any]
OptionalStrDict = Optional[Dict[str, str]]
OptionalFormData = Optional[Dict[str, Union[str, List[str]]]]
OptionalJson = Optional[Any]
OptionalBasicAuth = Optional[BasicAuth]


def _build_json_response(response: requests.Response) -> JsonResponse:
    try:
        response_json = response.json()
    except json.decoder.JSONDecodeError as e:
        logger.exception(e)
        logger.error("failed to parse response json")
        logger.error(f"response text: {response.text}")
        raise exceptions.ResponseParsingError(
            "The response is either invalid JSON or in a different format. Please view the logs for more information."
        )
    return JsonResponse(
        status_code=response.status_code, headers=dict(response.headers), cookies=dict(response.cookies), json=response_json
    )


def _build_base_kwargs(
    headers: OptionalStrDict, params: OptionalStrDict, cookies: OptionalStrDict, basic_auth: OptionalBasicAuth
) -> Kwargs:
    return {
        key: value_parser(value)
        for key, value, value_parser in [
            ("headers", headers, lambda x: x),
            ("params", params, lambda x: x),
            ("cookies", cookies, lambda x: x),
            ("auth", basic_auth, lambda x: (x["username"], x["password"])),
        ]
        if value is not None
    }


def _type_check_request_data(name, value, type_annotation):
    # This is purely here to check request data supplied using load in
    # the YAML, as the validation endpoint doesn't check those.
    try:
        check_type(name, value, type_annotation)
    except TypeError as e:
        logger.exception(e)
        raise exceptions.PostDataError(f"The supplied {name} is in an invalid format. Please view the logs for more information.")


def _build_request_data(form_data: OptionalJson, json_data: OptionalFormData) -> dict:
    if all([form_data is None, json_data is None]):
        return {}

    if all([form_data, json_data]):
        raise exceptions.RequestParameterError("Values were supplied for both form_data and json_data.")

    _type_check_request_data(*("form_data", form_data, OptionalFormData) if form_data else ("json_data", json_data, OptionalJson))

    return {"data" if form_data else "json": form_data or json_data}


def get(
    url: str,
    headers: OptionalStrDict = None,
    params: OptionalStrDict = None,
    cookies: OptionalStrDict = None,
    basic_auth: OptionalBasicAuth = None,
    verify: Optional[bool] = True,
) -> JsonResponse:
    """Make a GET request to a URL that returns a JSON response.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __url:__ the full url that you want to request

    - __headers:__ an optional dictionary with the request headers

    - __params:__ an optional dictionary with the query parameters

    - __cookies:__ an optional dictionary with the request cookies

    - __basic_auth:__ an optional [BasicAuth](../inputs_outputs/#BasicAuth) dictionary

    - __verify:__ set to `false` to skip certificate checking (defaults to `true`)

??? success "Output"
    A [JsonResponse](../inputs_outputs/#JsonResponse) dictionary.

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.

??? example
    Make an HTTP GET request using basic authentication.

    {% raw %}
    ```YAML linenums="1"
    - http_requests.json.get:
        url: https://httpbin.org/basic-auth/username/password
        basic_auth:
          username: username
          password: password
      save: response

    - ppa.ui.output_info:
        text: Response code was {{ response.status_code }}
    ```
    {% endraw %}
    """
    logger.info(f"sending GET request to '{url}'...")
    logger.debug(f"{headers=}")
    logger.debug(f"{cookies=}")
    response = requests.get(url, **_build_base_kwargs(headers, params, cookies, basic_auth), verify=verify)
    logger.info(f"response status code: {response.status_code}")
    logger.debug(response)
    logger.debug(response.text)
    return _build_json_response(response)


def post(
    url: str,
    headers: OptionalStrDict = None,
    params: OptionalStrDict = None,
    cookies: OptionalStrDict = None,
    basic_auth: OptionalBasicAuth = None,
    form_data: OptionalFormData = None,
    json_data: OptionalJson = None,
    verify: Optional[bool] = True,
) -> JsonResponse:
    """Make a POST request to a URL that returns a JSON response.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    - __url:__ the full url that you want to request

    - __headers:__ an optional dictionary with the request headers

    - __params:__ an optional dictionary with the query parameters

    - __cookies:__ an optional dictionary with the request cookies

    - __basic_auth:__ an optional [BasicAuth](../inputs_outputs/#BasicAuth) dictionary

    - __form_data:__ optional form data to send in the request

    - __json_data:__ optional JSON data to send in the request

    - __verify:__ set to `false` to skip certificate checking (defaults to `true`)

??? success "Output"
    A [JsonResponse](../inputs_outputs/#JsonResponse) dictionary.

!!! danger "Request Data"

    You cannot supply values for _both_ the `form_data` & `json_data` parameters to this action.

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.

??? example
    Making an HTTP POST request to fill in a form.

    {% raw %}
    ```YAML linenums="1"
    - http_requests.json.post:
        url: https://httpbin.org/post
        form_data:
          comments: ""
          custname: John Doe
          custemail: john.doe@fakemail.com
          custtel: "123456"
          delivery: "19:45"
          size: large
          topping:
            - bacon
            - cheese
      save: response

    - ppa.ui.output_info:
        text: Response code was {{ response.status_code }}
    ```
    {% endraw %}
    """
    logger.info(f"sending POST request to '{url}'...")
    logger.debug(f"{headers=}")
    logger.debug(f"{cookies=}")
    response = requests.post(
        url,
        **dict(_build_base_kwargs(headers, params, cookies, basic_auth), **_build_request_data(form_data, json_data)),
        verify=verify,
    )
    logger.info(f"response status code: {response.status_code}")
    logger.debug(response)
    logger.debug(response.text)
    return _build_json_response(response)


def patch(
    url: str,
    headers: OptionalStrDict = None,
    params: OptionalStrDict = None,
    cookies: OptionalStrDict = None,
    basic_auth: OptionalBasicAuth = None,
    form_data: OptionalFormData = None,
    json_data: OptionalJson = None,
    verify: Optional[bool] = True,
) -> JsonResponse:
    """Make a PATCH request to a URL that returns a JSON response.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    - __url:__ the full url that you want to request

    - __headers:__ an optional dictionary with the request headers

    - __params:__ an optional dictionary with the query parameters

    - __cookies:__ an optional dictionary with the request cookies

    - __basic_auth:__ an optional [BasicAuth](../inputs_outputs/#BasicAuth) dictionary

    - __form_data:__ optional form data to send in the request

    - __json_data:__ optional JSON data to send in the request

    - __verify:__ set to `false` to skip certificate checking (defaults to `true`)

??? success "Output"
    A [JsonResponse](../inputs_outputs/#JsonResponse) dictionary.

!!! danger "Request Data"

    You cannot supply values for _both_ the `form_data` & `json_data` parameters to this action.

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.

??? example
    Make an HTTP PATCH request with some query parameters.

    {% raw %}
    ```YAML linenums="1"
    - http_requests.json.patch:
        url: https://httpbin.org/patch
        params:
          key1: value1
          key2: value2
      save: response

    - ppa.ui.output_info:
        text: Response code was {{ response.status_code }}
    ```
    {% endraw %}
    """
    return _build_json_response(
        requests.patch(
            url,
            **dict(_build_base_kwargs(headers, params, cookies, basic_auth), **_build_request_data(form_data, json_data)),
            verify=verify,
        )
    )


def delete(
    url: str,
    headers: OptionalStrDict = None,
    params: OptionalStrDict = None,
    cookies: OptionalStrDict = None,
    basic_auth: OptionalBasicAuth = None,
    form_data: OptionalFormData = None,
    json_data: OptionalJson = None,
    verify: Optional[bool] = True,
) -> JsonResponse:
    """Make a DELETE request to a URL that returns a JSON response.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    - __url:__ the full url that you want to request

    - __headers:__ an optional dictionary with the request headers

    - __params:__ an optional dictionary with the query parameters

    - __cookies:__ an optional dictionary with the request cookies

    - __basic_auth:__ an optional [BasicAuth](../inputs_outputs/#BasicAuth) dictionary

    - __form_data:__ optional form data to send in the request

    - __json_data:__ optional JSON data to send in the request

    - __verify:__ set to `false` to skip certificate checking (defaults to `true`)

??? success "Output"
    A [JsonResponse](../inputs_outputs/#JsonResponse) dictionary.

!!! danger "Request Data"

    You cannot supply values for _both_ the `form_data` & `json_data` parameters to this action.

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.

??? example
    Make an HTTP DELETE request with some query parameters.

    {% raw %}
    ```YAML linenums="1"
    - http_requests.json.delete:
        url: https://httpbin.org/delete
        params:
          key1: value1
          key2: value2
      save: response

    - ppa.ui.output_info:
        text: Response code was {{ response.status_code }}
    ```
    {% endraw %}
    """
    return _build_json_response(
        requests.delete(
            url,
            **dict(_build_base_kwargs(headers, params, cookies, basic_auth), **_build_request_data(form_data, json_data)),
            verify=verify,
        )
    )
